(* file B2_SIF_pre_some_trad.ml *)
(* contains SKIBCN algorithm + S_IF combinator + combinators for multiple abstractions/applications *)
(* pre-evaluation of partial application of combinators when possible *)

open Term ;;
open Combinators ;;

exception UnexpectedError ;;


let rec skip x = function
  | P_Int _ -> true
  | P_Bool _ -> true
  | P_ExtFun _ -> true
  | P_Var y -> x <> y  (*false*) (* vérifier *) (* la version x<>y donne de meilleurs résultats mais est-elle correcte? *)
  | P_App (e1, e2) -> false
  | P_Abs (y, e) -> x=y || not (free x e) ;; 


let can_pre_eval = function
  | P_Int _ -> true
  | P_Bool _ -> true
  | P_ExtFun _ -> true
  | P_Var _ -> false
  | P_App _ -> false
  | P_Abs _ -> false

let pre_eval = function
  | P_Int i -> V_Int i
  | P_Bool b-> V_Bool b
  | P_ExtFun f -> V_Fun f
  | P_Var _ -> raise UnexpectedError
  | P_App _ -> raise UnexpectedError
  | P_Abs _ -> raise UnexpectedError

let valfun_to_extfun = function (V_Fun f) -> P_ExtFun f ;;

let rec pre_eval_app e = match e with
  | P_App(P_ExtFun f,a) -> 
      if can_pre_eval a 
      then valfun_to_extfun (f (pre_eval a)) 
      else e
  | P_App(g,a) -> 
      (match pre_eval_app g, can_pre_eval a with
	 | P_ExtFun g', true -> valfun_to_extfun (g' (pre_eval a))
	 | g'',_ -> P_App(g'', a)
      )
      

let rec elim = function
    P_Int i    -> P_Int i
  | P_Bool b   -> P_Bool b 
  | P_Var x    -> P_Var x
  | P_ExtFun f -> P_ExtFun f   

  | P_App (e1, e2) ->  P_App (elim e1, elim e2) 
      
  | P_Abs(s, P_Var s' ) -> 
      if s = s' 
      then  cI 
      else  (P_App (cK, P_Var s'))

  | P_Abs(_, P_Int i   ) -> P_ExtFun (fun _ -> V_Int i) 
  | P_Abs(_, P_Bool b  ) -> P_ExtFun (fun _ -> V_Bool b)
  | P_Abs(_, P_ExtFun f) -> P_ExtFun (fun _ -> V_Fun f)
 

(* case S_IF *)
  | P_Abs(s,P_App(P_App(P_App(P_ExtFun i,e1), P_Abs(_,e2)), P_Abs(_,e3))) when i  == comb_IF -> 
    P_App (P_App (P_App (cSIF, elim (P_Abs (s,e1))), elim (P_Abs(s,e2))), elim (P_Abs (s, e3)))
	(* see my PhD page 116/117 *)

(* case S_2_IF *)
  (* if this case is ommitted, the S_IF case might not be matched when the conditional is under two lambdas *)
  | P_Abs(x,P_Abs(y,P_App(P_App(P_App(P_ExtFun i,e1), P_Abs(_,e2)), P_Abs(_,e3)))) when i  == comb_IF ->
      let e1' = P_Abs(x,P_Abs (y,e1))
      and e2' = P_Abs(x,P_Abs (y,e2)) 
      and e3' = P_Abs(x,P_Abs (y,e3)) in
	P_App (P_App (P_App (cS_2_IF, elim e1'), elim e2'), elim e3')
(* FIXME : B_IF and C_IF and N_IF would be nice too *)


(* case S_2_2 *)
  | P_Abs(x, P_Abs (y, P_App (P_App(e1,e2), e3))) ->
      let e1' = P_Abs (x, P_Abs(y, e1)) 
      and e2' = P_Abs (x, P_Abs(y, e2))  
      and e3' = P_Abs (x, P_Abs(y, e3))  
      in pre_eval_app (P_App (P_App (P_App (cS_2_2,  elim e1'), elim e2' ), elim e3'))
	   

  

(* case S_2_1 *)

  | P_Abs(x, P_Abs (y, P_App (e1,e2))) ->
      let e1' = P_Abs (x, P_Abs(y, e1)) 
      and e2' = P_Abs (x, P_Abs(y, e2))  
      in pre_eval_app (P_App( P_App(cS_2_1, elim e1'), elim e2'))

	   
(* case S_1_2 *)
  | P_Abs (x,P_App (P_App (e1,e2), e3)) ->  
      
      let e1' = P_Abs (x, e1) 
      and e2' = P_Abs (x, e2) 
      and e3' = P_Abs (x, e3) 
      in 
      let (c, a1, a2, a3) = 
	(match (skip x e1, skip x e2, skip x e3) with
	   | (true , true , true ) -> (cNNN, e1  , e2  , e3  )
	   | (true , true , false) -> (cNNY, e1  , e2  , e3' )
	   | (true , false, true ) -> (cNYN, e1  , e2' , e3  )
	   | (true , false, false) -> (cNYY, e1  , e2' , e3' )
	   | (false, true , true ) -> (cYNN, e1' , e2  , e3  )
	   | (false, true , false) -> (cYNY, e1' , e2  , e3' )
 	   | (false, false, true ) -> (cYYN, e1' , e2' , e3  )
	   | (false, false, false) -> (cYYY, e1' , e2' , e3' )
	)
      in pre_eval_app(P_App(P_App(P_App(c, elim a1), elim a2), elim a3))
	   
	   
  
(* case B, C, N, S *) 

  | P_Abs(s,P_App(e1,e2)) ->
      (match (skip s e1,skip s e2) with
	 | true,  true  -> pre_eval_app( P_App(P_App(cNN, elim e1)           , elim e2)            )
	 | true,  false -> pre_eval_app( P_App(P_App(cNY, elim e1)           , elim (P_Abs(s,e2))) )
	 | false, true  -> pre_eval_app( P_App(P_App(cYN, elim (P_Abs(s,e1))), elim e2)            )
	 | false, false -> pre_eval_app( P_App(P_App(cYY, elim (P_Abs(s,e1))), elim (P_Abs(s,e2))) )
      )
 
  | P_Abs(s, (P_Abs(_,_) as t))-> 
      
       elim (P_Abs(s,elim t)) 

  ;; 
 
let trans e = check_no_var (elim e) ;;
