(* file BCN_SIF_2_trad.ml *)
(* with S B C N K I and S_IF and S_Cons combinators *)
(* skip the lambda distribution only when no application in a branch *)
(* some partial applications of combinators are done *)

open Term 
open Combinators 



let rec skip x = function
  | P_Int _ -> true
  | P_Bool _ -> true
  | P_ExtFun _ -> true
  | P_Var y -> x <> y  (*false*) (* vérifier *) (* la version x<>y donne de meilleurs résultats mais est-elle correcte? *)
  | P_App (e1, e2) -> false
  | P_Abs (y, e) ->  raise UnexpectedError;;

let can_pre_eval = function
  | P_Int _ -> true
  | P_Bool _ -> true
  | P_ExtFun _ -> true
  | P_Var _ -> false
  | P_App _ -> false
  | P_Abs _ -> false

let pre_eval = function
  | P_Int i -> V_Int i
  | P_Bool b-> V_Bool b
  | P_ExtFun f -> V_Fun f
  | P_Var _ -> raise UnexpectedError
  | P_App _ -> raise UnexpectedError
  | P_Abs _ -> raise UnexpectedError

(* the type cannot change since ski is applied to its result *)
let rec bcn = function
    P_Int i    -> P_Int i
  | P_Bool b   -> P_Bool b 
  | P_Var x    -> P_Var x
  | P_ExtFun f -> P_ExtFun f   

  | P_App (e1, e2) ->  P_App (bcn e1, bcn e2) 
      
  | P_Abs(s, P_Var s' ) -> 
      if s = s' 
      then  cI 
      else  (P_App (cK, P_Var s'))

  | P_Abs(_, P_Int i   ) -> P_ExtFun (fun _ -> V_Int i) 
  | P_Abs(_, P_Bool b  ) -> P_ExtFun (fun _ -> V_Bool b)
  | P_Abs(_, P_ExtFun f) -> P_ExtFun (fun _ -> V_Fun f)
 
  | P_Abs(s,P_App(P_App(P_App(P_ExtFun i,e1), P_Abs(_,e2)), P_Abs(_,e3))) when i  == comb_IF ->
      P_App (P_App (P_App (cSIF, bcn (P_Abs (s,e1))), bcn (P_Abs(s,e2))), bcn (P_Abs (s, e3)))
	(* see my PhD page 116/117 *)
(* manque la pré-évaluation *)



(* ------- combinator for cons : try 1  ----------- *)
(*
  | P_Abs(s,P_App(P_App(P_ExtFun i,e1), e2)) when i == comb_Cons ->
      ( let (r1,r2) = (bcn e1, bcn e2) in 
	  match (can_pre_eval r1, skip s r1, can_pre_eval r2, skip s r2) with					       
	    | (_, true, _, true)   -> P_App(P_App(cConsNN, r1)               ,r2)
	    | (_, true, _, false)  -> P_App(P_App(cConsNY, r1)               , bcn (P_Abs(s,r2)))
	    | (_, false, _, true)  -> P_App(P_App(cConsYN, bcn (P_Abs(s,r1))), r2)
	    | (_, false, _, false) -> P_App(P_App(cConsYY, bcn (P_Abs(s,r1))), bcn (P_Abs(s,r2)))
		
      )*)

(* note : reecrire BCN2 comme ci-dessus *)
(* et voir si il faut appliquer BCN avant ou après avoir regardé skippable et pre_evaluable *)


(* ------- combinator for cons : try 2  ----------- *)


(* For S_Cons the gain can not be as great as the one with S_IF, because S_IF allows to eliminate one/two thunks (dummy lambdas) which S_COMB does not. *)
  | P_Abs(s,P_App(P_App(P_ExtFun i,e1), e2)) when i == comb_Cons ->  
      ( let (r1,r2) = (bcn e1, bcn e2) in 
	  match (can_pre_eval r1, skip s r1, can_pre_eval r2, skip s r2) with 
	    | (true, _, true, _) -> let v= V_Cons(pre_eval r1,pre_eval r2) in (P_ExtFun (fun _ -> v)) 
										
	    | (true, _, false, true)  -> (* e2 is a free variable *) let v = pre_eval r1 in P_App (P_ExtFun (fun x -> V_Fun (fun _ -> V_Cons(v,x))), r2) 
	    | (true, _, false, false) -> let v = pre_eval r1 in P_App(P_ExtFun (fun (V_Fun f) -> V_Fun (fun x -> V_Cons(v, f x))), bcn (P_Abs(s,r2))) 
								  
	    | (false, true, true,  _)      -> (*e1 is a free variable *) let v = pre_eval r2 in P_App (P_ExtFun (fun x -> V_Fun (fun _ -> V_Cons (x,v))), r1) 
	    | (false, false, true, _)      -> let v = pre_eval r2 in P_App(P_ExtFun (fun (V_Fun f) -> V_Fun (fun x -> V_Cons(f x, v))), bcn (P_Abs(s,r1))) 
								       
	    | (false, true, false, true)   -> P_App(P_App(cConsNN, r1)               ,r2) 
	    | (false, true, false, false)  -> P_App(P_App(cConsNY, r1)               , bcn (P_Abs(s,r2))) 
	    | (false, false, false, true)  -> P_App(P_App(cConsYN, bcn (P_Abs(s,r1))), r2)
	    | (false, false, false, false) -> P_App(P_App(cConsYY, bcn (P_Abs(s,r1))), bcn (P_Abs(s,r2))) 
		 
      )

      



(* case for BCN and variants with pre-evaluation *)
  | P_Abs(s,P_App(e1,e2)) ->
      let (r1,r2) = (bcn e1, bcn e2) in
	(match (can_pre_eval r1, can_pre_eval r2) with

	   | (true,true) ->
	       let (V_Fun f,v) = (pre_eval r1, pre_eval r2) in
		 P_ExtFun (fun _ -> f v) (* correct because can_pre_eval reports that r1 and r2 are closed *)
		   
	   | (true, false) ->
	       let (V_Fun f) = pre_eval r1 in
		 (match skip s r2 with
		    | true -> (* this case is the difference between can_pre_eval and skip, that is free variable. *) (* NN is used as a model *)
			P_App (P_ExtFun (fun e -> V_Fun (fun _ -> f e)), r2) 
		    | false -> (* NY is used as a model *)
			P_App (P_ExtFun (fun (V_Fun g) -> V_Fun (fun x -> f (g x))), bcn (P_Abs(s,r2))) 

		 )

	   | (false,true) ->
	       (* r1 cannot be pre-evaluated but r2 can (and skip r2 is surely true) *)
	       let v = pre_eval r2 in
		 (
		   match skip s r1 with
		     | true -> (* r1 is surely a variable *)
			 (* NN is used as a model *)
			 P_App (P_ExtFun ( fun (V_Fun f) -> V_Fun (fun _ -> f v)) , bcn (P_Abs(s,r1)))
		     
		     | false ->
			 (* YN is used as a model *)
			 P_App(P_ExtFun (fun (V_Fun f) ->
		       
				 V_Fun ( fun x ->
					   match  (f x) with V_Fun (h) -> (* caml application *)
					   h v )), bcn (P_Abs(s,r1)))
	       )

	   | (false,_) ->
	       (match (skip s r1,skip s r2) with
		  | true , true  -> P_App(P_App(cNN, r1)               ,r2)
		  | true , false -> P_App(P_App(cNY, r1)               , bcn (P_Abs(s,r2)))
		  | false, true  -> P_App(P_App(cYN, bcn (P_Abs(s,r1))), r2)
		  | false, false -> P_App(P_App(cYY, bcn (P_Abs(s,r1))), bcn (P_Abs(s,r2)))
	       )

	)
  | P_Abs(s, (P_Abs(_,_) as t))->
      
       bcn (P_Abs(s,bcn t))

  ;;

let trans e = check_no_var (bcn e) ;;

