(* file BCN_SIF_premix_trad.ml *)
(* with S B C N K I and S_IF combinators *)
(* skip the lambda distribution only when no application in a branch (BCN) *)
(* some partial applications of combinators are done, automatically and by hand *)


open Term ;;

(* ---------------------   combinators  ---------------------------- *)

(* combinators S, K and I *)


let cI = P_Const (V_Fun (fun x -> x)) 
and cK = P_Const (V_Fun (fun c -> V_Fun (fun _ -> c)) )
and cS = P_Const (V_Fun ( 
  fun (V_Fun f) -> V_Fun (
    fun (V_Fun g) -> V_Fun ( 
      fun x -> match (f x) with     (* 1 caml capplication *) 
          V_Fun h -> h (g x)     )))) (* 2 caml applications *) 
;;

(* combinator S_IF *)

let cSIF =  P_Const (V_Fun (fun (V_Fun e1) -> 
                       V_Fun (fun (V_Fun e2) ->
                                            V_Fun (fun (V_Fun e3) ->
                                                     V_Fun (fun x ->
  match e1 x with (* caml application *)
      V_Bool true  -> e2 x  (* caml application *)
    | V_Bool false -> e3 x  (* caml application *)
    | _ -> raise TypeError                        )))))
;;

(* combinators B C N *)

let cYN = P_Const (V_Fun (fun (V_Fun f) ->
		       V_Fun (fun e ->
				 V_Fun ( fun x ->
					   match  (f x) with V_Fun (h) -> (* caml application *)
					   h e ))));;                      (* caml application *)

let cNY = P_Const (V_Fun (fun (V_Fun f) ->
		       V_Fun (fun (V_Fun (g)) ->
				 V_Fun ( fun x -> f (g x) ))));; (* 2 caml applications *)

let cNN =  P_Const (V_Fun (fun (V_Fun f) ->
		       V_Fun (fun e ->
				 V_Fun ( fun _ ->
					   f e)))) ;; (* caml application *)
let cYY = cS ;;


(* ----------------------------------------------------------------- *)



let rec skip x = function
  | P_Const _ -> true
  | P_Var y -> x <> y  (*false*) (* vérifier *) (* la version x<>y donne de meilleurs résultats mais est-elle correcte? *)
  | P_App (e1, e2) -> false
  | P_Abs (y, e) -> x=y || not (free x e) ;;

let can_pre_eval = function
  | P_Const _ -> true
  | P_Var _ -> false
  | P_App _ -> false
  | P_Abs _ -> false

let pre_eval = function
  | P_Const c -> c
  | P_Var _ -> raise UnexpectedError
  | P_App _ -> raise UnexpectedError
  | P_Abs _ -> raise UnexpectedError


let rec elim = function

  | P_Const c -> P_Const c
  | P_Var x    -> P_Var x

  | P_App (e1, e2) ->  P_App (elim e1, elim e2) 

(* -- Case -- K *)
  | P_Abs(s, P_Var s' ) -> 
      if s = s' 
      then  cI 
      else  (P_App (cK, P_Var s'))

  | P_Abs(_,P_Const c) -> P_Const (V_Fun (fun _ -> c)) (* hand-made pre-eval *)

 (* case S_IF *)
  | P_Abs(s,P_App(P_App(P_App(i,e1), P_Abs(_,e2)), P_Abs(_,e3))) when i  == Purify.cIF ->
      P_App (P_App (P_App (cSIF , elim (P_Abs (s,e1))), elim (P_Abs(s,e2))), elim (P_Abs (s, e3)))
	(* see my PhD page 116/117 *)

(* FIXME : B_IF and C_IF and N_IF would be nice too *)

  | P_Abs(s,P_App(e1,e2)) ->
      let (r1,r2) = (elim e1, elim e2) in
	(match (can_pre_eval r1, can_pre_eval r2) with

	   | (true,true) ->
	       let (V_Fun f,v) = (pre_eval r1, pre_eval r2) in
		 P_Const (V_Fun (fun _ -> f v)) (* correct because can_pre_eval reports that r1 and r2 are closed *)
		   
	   | (true, false) ->
	       let (V_Fun f) = pre_eval r1 in
		 (match skip s r2 with
		    | true -> (* this case is the difference between can_pre_eval and skip, that is free variable. *) (* NN is used as a model *)
			P_App (P_Const (V_Fun (fun e -> V_Fun (fun _ -> f e))), r2) 
		    | false -> (* NY is used as a model *)
			P_App (P_Const (V_Fun (fun (V_Fun g) -> V_Fun (fun x -> f (g x)))), elim (P_Abs(s,r2))) 

		 )

	   | (false,true) ->
	       (* r1 cannot be pre-evaluated but r2 can (and skip r2 is surely true) *)
	       let v = pre_eval r2 in
		 (
		   match skip s r1 with
		     | true -> (* r1 is surely a variable *)
			 (* NN is used as a model *)
			 P_App (P_Const (V_Fun ( fun (V_Fun f) -> V_Fun (fun _ -> f v))) , elim (P_Abs(s,r1)))
		     
		     | false ->
			 (* YN is used as a model *)
			 P_App(P_Const (V_Fun (fun (V_Fun f) ->
		       
				 V_Fun ( fun x ->
					   match  (f x) with V_Fun (h) -> 
					   h v ))), elim (P_Abs(s,r1)))
	       )

	   | (false,_) ->
	       (match (skip s r1,skip s r2) with
		  | true, true ->
		      P_App(P_App(cNN , r1),r2)
		  | true, false ->
		      P_App(P_App(cNY , r1), elim (P_Abs(s,r2)))
		  | false, true ->
		      P_App(P_App(cYN, elim (P_Abs(s,r1))), r2)
		  | false , false -> 
		      P_App(P_App(cYY , elim (P_Abs(s,r1))), elim (P_Abs(s,r2)))
	       )
		 
	)
  | P_Abs(s, (P_Abs(_,_) as t))->
      
       elim (P_Abs(s,elim t))

  ;;

let trans e = check_no_var (elim e) ;;
