type ident = string ;;


type 'a environment = (ident * 'a) list 
;;


(* The type of values returned by the evaluation function *)
type value = 
  | V_Int of int
  | V_Bool of bool
  (* no Var *)
  (* no Abs *)
  (* no App *)
  | V_ExtFun of (value->value)
  (* no If *)
  | V_Closure of ident * expr * (value environment) 

  | V_List of value list

and expr = 
  | E_Const of value
  | E_Var of ident
  | E_Abs of ident * expr
  | E_App of expr * expr
  | E_If of expr * expr * expr

  | E_Get of Store.tag
  | E_Set of Store.tag * expr


exception TypeError ;;


let rec string_of_value = function
  | V_Int i -> string_of_int i
  | V_Bool true -> "true"
  | V_Bool false -> "false"
  | V_ExtFun _ -> "ext"
  | V_Closure _ -> "fun" 

  | V_List l -> "[ " ^ List.fold_right (fun x-> fun r -> string_of_value x ^ " , " ^ r) l " ]"

;;


(*
let rec freevars l = function
  | E_Const _ -> []
  | E_App (e1,e2) -> freevars l e1 @ freevars l e2
  | E_Var x -> if List.mem x l then [] else [x]
  | E_Set _ -> []
  | E_Get _ -> []
  | E_Abs (x,e) -> freevars (x::l) e
  | E_If (a,b,c) -> freevars l a @ freevars l b @ freevars l c
*)

let rec freevars l = function
  | E_Const _ -> false
  | E_App (e1,e2) -> freevars l e1 || freevars l e2
  | E_Var x -> not (List.mem x l )
  | E_Set (_,e) -> freevars l e
  | E_Get _ -> false
  | E_Abs (x,e) -> freevars (x::l) e
  | E_If (a,b,c) -> freevars l a || freevars l b || freevars l c


(* -- The classical evaluation function with environments -- *)
let (store : value Store.store ref) = ref [] ;;

let rec eval env = function
  | E_Const c -> c
  | E_Var x -> List.assoc x env 
  | E_Abs (x,e) -> V_Closure (x,e,env)

(*  | E_App (e1,e2) ->
      (match (eval env e1, eval env e2) with 
         | V_Closure (x,e,env2), v2 -> eval ( (x,v2) :: env2 ) e
         | V_ExtFun f , v2-> f v2  (* caml application here *)
         | _ -> raise TypeError                              
        ) 
*)
     | E_App (e1,e2) ->
      (match (eval env e1) with 
         | V_Closure (x,e,env2)-> eval ( (x,eval env e2) :: env2 ) e
         | V_ExtFun f -> f (eval env e2)  (* caml application here *)
         | _ -> raise TypeError                              
        )       
  | E_If (e1,e2,e3) ->
      (match eval env e1 with
         | V_Bool true  -> eval env e2
         | V_Bool false -> eval env  e3
         | _ -> raise TypeError                                 )
	
(*  | E_Cons (e1,e2) ->
      let v2 = eval env e2 in
      let v1 = eval env e1 in
      V_Cons (v1, v2)
*)

  | E_Get t -> Store.get t !store 
  | E_Set (t,e) -> 
      let v = eval env e in 
        begin store := Store.set t v !store ; v end  

;;


