open Term ;;

exception TypeError ;;

let dummy_val = V_Bool true ;;

(*let comb_IF =  
  fun e1 ->  V_Fun (
    fun (V_Fun e2) -> V_Fun (
      fun (V_Fun e3) ->
        match e1 with
            V_Bool true  -> e2 dummy_val (* caml application *)
          | V_Bool false -> e3 dummy_val (* caml application *)
          | _ -> raise TypeError                             ))
;;
*)

(* version code factorisé + unwrap *)
(* 
let comb_IF =  fun e1 ->  V_Fun (
   fun (e_then) -> V_Fun (
     fun (e_else) -> (unwrap (match e1 with
           V_Bool true  -> e_then
         | V_Bool false -> e_else
         | _ -> raise TypeError                             ) )
   dummy_val ))
;;*)


(* version code factorisé sans unwrap *)

let comb_IF =  fun e1 ->  V_Fun (
   fun (e_then) -> V_Fun (
     fun (e_else) -> (match 
			(match e1 with
			     V_Bool true  -> e_then
			   | V_Bool false -> e_else
			   | _ -> raise TypeError     )
		      with V_Fun f ->  f  dummy_val )))
;;



(* ---------------------------------- External functions ------------------------------------------- *)

let print2 (v1,v2) = let s = string_of_value v1 ^ " / " ^ string_of_value v2 in print_endline s ;;

let wrap_int_int_int f =  
  (fun x -> V_Fun (fun y -> 
	 match (x,y) with 
	   | V_Int a, V_Int b -> V_Int (f a b)
	   | v1,v2 -> (*print2 (v1,v2) ;*) raise TypeError            ))
;;

let wrap_bool_bool_bool f = 
  (fun x -> V_Fun (fun y -> 
	 match (x,y) with 
	   | V_Bool a, V_Bool b -> V_Bool (f a b)
	   | v1,v2 -> (*print2 (v1,v2) ;*)  raise TypeError            ))
;;

let wrap_int_int_bool f = 
  (fun x -> V_Fun (fun y -> 
	 match (x,y) with 
	   | V_Int a, V_Int b -> V_Bool (f a b)
	   | v1,v2 -> (*print2 (v1,v2) ;*)  raise TypeError            ))
;;

let plus   = wrap_int_int_int (+) ;;
let minus  = wrap_int_int_int (-) ;;
let inf    = wrap_int_int_bool (<=) ;;
let equals = wrap_int_int_bool (=) ;;
let diff  = wrap_int_int_bool (<>) ;;
let log_and = wrap_bool_bool_bool (&&) ;;

let vtrue  = V_Bool true
and vfalse = V_Bool false

(*let not_op = P_ExtFun (function V_Bool true -> vfalse | V_Bool false -> vtrue )
*)

let wrap_list_list f = fun (V_List a) -> V_List (f a) ;;


let ccons = fun x -> V_Fun (fun (V_List l) -> V_List (x :: l)) ;;

let empty = function
  | V_List [] -> vtrue
  | V_List _  -> vfalse

let hd = function 
  | V_List (a::_) -> a

let tl = wrap_list_list List.tl
let rev = wrap_list_list List.rev

let flatten = wrap_list_list (fun l -> List.flatten (List.map (fun (V_List l) -> l) l))


let print v = begin print_endline (string_of_value v) ; V_Int 0 end ;;

let length = function (V_List l) -> V_Int (List.length l) ;;

let iter = function (V_Fun f) -> 
  V_Fun (function (V_List l) -> 
	   begin List.iter (fun v -> begin f v ; () end ) l ; V_Int 0 end ) ;;

let filter = function (V_Fun f) ->
  V_Fun (function (V_List l) ->
	   V_List (List.filter (function v -> (match f v with V_Bool t -> t)) l)) 
    

let host_compose f g  = fun x -> f (g x) ;;

let compose_embed = 
  V_Fun (function (V_Fun f) -> 
	   V_Fun (function (V_Fun g) ->
		    V_Fun (host_compose f g) )) ;;



(* ------------------- Imperative combinators --------------------------- *)


let (store : value Store.store ref) = ref [] ;;

let make_get t = 
 P_Const (V_Fun (fun _ -> Store.get t !store))
 (* peut-on faire mieux en applicant partiellement assoc? non *)

let make_set t = P_Const (V_Fun (fun v -> begin  store := Store.set t v !store ; v end))
