open Classical ;;

exception TypeError ;;

let dummy_val = V_Bool true ;;


(* ---------------------------------- External functions ------------------------------------------- *)

let print2 (v1,v2) = let s = string_of_value v1 ^ " / " ^ string_of_value v2 in print_endline s ;;

let wrap_int_int_int f =  
  (fun x -> V_ExtFun (fun y -> 
	 match (x,y) with 
	   | V_Int a, V_Int b -> V_Int (f a b)
	   | v1,v2 -> print2 (v1,v2) ; raise TypeError            ))
;;

let wrap_bool_bool_bool f = 
  (fun x -> V_ExtFun (fun y -> 
	 match (x,y) with 
	   | V_Bool a, V_Bool b -> V_Bool (f a b)
	   | v1,v2 -> print2 (v1,v2) ;  raise TypeError            ))
;;

let wrap_int_int_bool f = 
  (fun x -> V_ExtFun (fun y -> 
	 match (x,y) with 
	   | V_Int a, V_Int b -> V_Bool (f a b)
	   | v1,v2 -> print2 (v1,v2) ;  raise TypeError            ))
;;

let plus   = wrap_int_int_int (+) ;;
let minus  = wrap_int_int_int (-) ;;
let inf    = wrap_int_int_bool (<=) ;;
let equals = wrap_int_int_bool (=) ;;
let diff  = wrap_int_int_bool (<>) ;;
let log_and = wrap_bool_bool_bool (&&) ;;

let vtrue  = V_Bool true
and vfalse = V_Bool false

(*let not_op = P_ExtFun (function V_Bool true -> vfalse | V_Bool false -> vtrue )
*)

let wrap_list_list f = fun (V_List a) -> V_List (f a) ;;


let ccons = fun x -> V_ExtFun (fun (V_List l) -> V_List (x :: l)) ;;

let empty = function
  | V_List [] -> vtrue
  | V_List _  -> vfalse

let hd = function 
  | V_List (a::_) -> a

let tl = wrap_list_list List.tl
let rev = wrap_list_list List.rev

let flatten = wrap_list_list (fun l -> List.flatten (List.map (fun (V_List l) -> l) l))


let print v = begin print_endline (string_of_value v) ; V_Int 0 end ;;

let length = function (V_List l) -> V_Int (List.length l) ;;

(* iter and filter are higher-order functions. 

We want to reuse iter and filter from the standard OCaml library,
 we do not want to rewrite them.

Higher-order functions from the OCaml standard library receive OCaml functions.
So we have to transform closures into OCaml functions.
This is done with the app which behaves as a part of the eval function. *)

let app f v = match f with
  | V_ExtFun h -> h v 
  | V_Closure (x,e,env) -> eval ( (x,v) :: env ) e
;;
      
let iter = function f -> 
  V_ExtFun (function (V_List l) -> 
	      begin List.iter (fun v -> begin app f v ; () end ) l ; V_Int 0 end ) ;;

let filter = function f ->
  V_ExtFun (function (V_List l) ->
	      V_List (List.filter (function v -> (match app f v with V_Bool t -> t))  l)
		
	   ) 
(*
let iter = function f -> 
  V_ExtFun (function (V_List l) -> 
	      begin List.iter (fun v -> begin ( match f with
						  | V_ExtFun h -> h v 
						  | V_Closure (x,e,env) -> eval ( (x,v) :: env ) e
					      ) ; () end ) l ; V_Int 0 end ) ;;


let filter = function f ->
  V_ExtFun (function (V_List l) ->
	      V_List (List.filter (function v -> (let V_Bool t =
						    ( match f with
							| V_ExtFun h -> h v 
							| V_Closure (x,e,env) -> eval ( (x,v) :: env ) e
						    ) in t))  l)
		
	   ) 
*)


(* c'est la galere (f doit être une expr) *)




(* --- *)
(* ESSAI POUR ARTICLE *)
let compose f g x = f (g x)

(* the simplest way is the following but this is not what we want to do *)
(* function f -> V_Fun (function g -> V_Fun (function x -> (app f (app g x))) *)

let compose_embed = 
  V_ExtFun (function f -> 
	      V_ExtFun (function g -> 
			  V_ExtFun (compose (app f) (app g) )))
