open Term ;;
open Externals ;;

let dummy_var  = "_" ;;
let dummy_val  = V_Int 0 ;;
let dummy_expr = P_Const dummy_val ;;

let fun_IF =  
  fun e1 ->  V_Fun (
    fun (V_Fun e2) -> V_Fun (
      fun (V_Fun e3) ->
        match e1 with
          | V_Bool true  -> e2 dummy_val (* caml application *)
          | V_Bool false -> e3 dummy_val (* caml application *)
          | _ -> raise TypeError                             ))
;;

let cIF = P_Const (V_Fun fun_IF) ;;


let rec purify = function 
  | E_Const c -> P_Const c
  | E_Var x -> P_Var x
  | E_Abs (x,e) -> P_Abs (x, purify e)
  | E_App (e1, e2) -> P_App (purify e1, purify e2)
  | E_If (e1, e2, e3) ->  
      let e'1 = purify e1
      and e'2 = P_Abs(dummy_var, purify e2)
      and e'3 = P_Abs(dummy_var, purify e3)
      in P_App(P_App(P_App (cIF,e'1),e'2),e'3) 
      



(* Note, if e2 or e3 do not contain applications, no need to suspend them. *)


   | E_Get m -> P_App (make_get m , dummy_expr)
   | E_Set (m,e) -> P_App (make_set m, purify e) 
;;

(* Moreover, if they can be pre-evaluated,
   they can be embedded in a special IF combinator
   (as is done in SKI2 and BCN2).                                          *)
