(* file S2_K2_SIF_trad.ml *)
(* contains SKI algorithm + S_IF combinator + S_2_2 S_2_1 and S_1_2 combinators + K_2 combinator *)

open Term ;;
open Combinators ;;



let rec ski_s2 = function
    P_Int i    -> P_Int i
  | P_Bool b   -> P_Bool b 
  | P_Var x    -> P_Var x
  | P_ExtFun f -> P_ExtFun f   

  | P_App (e1, e2) ->  P_App (ski_s2 e1, ski_s2 e2) 
      
  | P_Abs(s, P_Var s' ) -> 
      if s = s' 
      then  cI 
      else  (P_App (cK, P_Var s'))

  | P_Abs(_, P_Int i   ) -> P_App (cK, P_Int i ) 
  | P_Abs(_, P_Bool b  ) -> P_App (cK, P_Bool b)
  | P_Abs(_, P_ExtFun f) -> P_App (cK, P_ExtFun f )

  | P_Abs(_, P_Abs(_, P_Int i)) -> P_App(cK_2, P_Int i)
  | P_Abs(_, P_Abs(_, P_Bool b)) -> P_App(cK_2, P_Bool b)
  | P_Abs(_, P_Abs(_, P_ExtFun f)) -> P_App(cK_2, P_ExtFun f)

  | P_Abs(x, P_Abs(y, P_Var z)) when (x<>z && y<>z)-> P_App(cK_2, P_Var z)

(* case S_IF *)
  | P_Abs(s,P_App(P_App(P_App(P_ExtFun i,e1), P_Abs(_,e2)), P_Abs(_,e3))) when i  == comb_IF ->
      P_App (P_App (P_App (cSIF, ski_s2 (P_Abs (s,e1))), ski_s2 (P_Abs(s,e2))), ski_s2 (P_Abs (s, e3)))
	(* see my PhD page 116/117 *)

(* case S_2_IF *)
(* if this case is ommitted, the S_IF case might not be matched when the conditional is under two lambdas *)
  | P_Abs(x,P_Abs(y,P_App(P_App(P_App(P_ExtFun i,e1), P_Abs(_,e2)), P_Abs(_,e3)))) when i  == comb_IF ->
      let e1' = P_Abs(x,P_Abs (y,e1))
      and e2' = P_Abs(x,P_Abs (y,e2))
      and e3' = P_Abs(x,P_Abs (y,e3)) in
      P_App (P_App (P_App (cS_2_IF, ski_s2 e1'), ski_s2 e2'), ski_s2 e3')

(* case S_2_2 *)
  | P_Abs(x, P_Abs (y, P_App (P_App(e1,e2), e3))) -> 
      let e1' = P_Abs (x, P_Abs(y, e1)) 
      and e2' = P_Abs (x, P_Abs(y, e2)) 
      and e3' = P_Abs (x, P_Abs(y, e3)) 
      in P_App (P_App (P_App (cS_2_2,  ski_s2 e1'), ski_s2 e2' ), ski_s2 e3')
	   
(* case S_2_1 *)
  | P_Abs(x, P_Abs (y, P_App (e1,e2))) ->
      let e1' = P_Abs (x, P_Abs(y, e1))
      and e2' = P_Abs (x, P_Abs(y, e2))
      in P_App( P_App(cS_2_1, ski_s2 e1'), ski_s2 e2')
	   
(* case S_1_2 *)
  | P_Abs (x,P_App (P_App (e1,e2), e3)) ->
      let e1' = P_Abs (x, e1) 
      and e2' = P_Abs (x, e2) 
      and e3' = P_Abs (x, e3) 
      in P_App (P_App (P_App (cS_1_2,  ski_s2 e1'), ski_s2 e2' ), ski_s2 e3')
    
(* case S_1_1 = case S *)
  | P_Abs(s,P_App(e1,e2)) ->
       P_App(P_App(cS, ski_s2 (P_Abs(s,e1))), ski_s2 (P_Abs(s,e2)))
      
  | P_Abs(s, (P_Abs(_,_) as t))->
      
       ski_s2 (P_Abs(s,ski_s2 t))

  ;;

let trans e = check_no_var (ski_s2 e) ;;
