(* file S2_SIF_trad.ml *)
(* Contains SKI algorithm + S_IF combinator + S_2_2 S_2_1 and S_1_2 combinators + S2IF *)
open Term ;;

(* ---------------------   combinators  ---------------------------- *)

(* combinators S, K and I *)


let cI = P_Const (V_Fun (fun x -> x)) 
and cK = P_Const (V_Fun (fun c -> V_Fun (fun _ -> c)) )
and cS = P_Const (V_Fun ( 
  fun (V_Fun f) -> V_Fun (
    fun (V_Fun g) -> V_Fun ( 
      fun x -> match (f x) with     (* 1 caml capplication *) 
          V_Fun h -> h (g x)     )))) (* 2 caml applications *) 
;;

(* combinator S_IF *)

let cSIF =  P_Const (V_Fun (fun (V_Fun e1) -> 
                       V_Fun (fun (V_Fun e2) ->
                                            V_Fun (fun (V_Fun e3) ->
                                                     V_Fun (fun x ->
  match e1 x with (* caml application *)
      V_Bool true  -> e2 x  (* caml application *)
    | V_Bool false -> e3 x  (* caml application *)
    | _ -> raise TypeError                        )))))
;;
(* ---------------------- Combinators S_2_2, S_2_1 and S_1_2 --------------------------------------- *)

(* two abstractions on two applications *)
let cS_2_2 = P_Const (V_Fun (fun (V_Fun f) -> 
		   V_Fun (fun (V_Fun g) ->
			    V_Fun (fun (V_Fun h) ->
				     V_Fun (fun x ->
					      V_Fun (fun y -> 
						       let (V_Fun f', V_Fun g', V_Fun h') =  (f x, g x, h x) in
						       let (V_Fun f'', g'', h'') = (f' y, g' y, h' y) in
						       let (V_Fun f''') = f'' g'' in f''' h'')
					   )))) )
;;

(* two abstractions on one application *)
let cS_2_1 = P_Const (V_Fun (fun (V_Fun f) -> 
		   V_Fun (fun (V_Fun g) ->
			    V_Fun (fun x ->
				     V_Fun (fun y -> 
					      let (V_Fun f', V_Fun g') =  (f x, g x) in
					      let (V_Fun f'', g'') = (f' y, g' y) in
					      f'' g'')
				  ))) )
;;

        
(* one abstraction on two applications *)
let cS_1_2 = P_Const (V_Fun (fun (V_Fun f) -> 
                   V_Fun (fun (V_Fun g) ->
                            V_Fun (fun (V_Fun h) ->
                                     V_Fun (fun x ->
                                              let (V_Fun f', g', h') =  (f x, g x, h x) in
                                              let (V_Fun f'') = f' g' in f'' h')
                                  ))) )
;;
(* two abstractions on a conditional *)
										    
let cS_2_IF =  P_Const (V_Fun ((fun (V_Fun e1) ->  
                       V_Fun (fun (V_Fun e2) ->
                                            V_Fun (fun (V_Fun e3) ->
                                                     V_Fun (fun x -> 
							      V_Fun (fun y ->
								       let (V_Fun f) = e1 x in match f y with 
									 | V_Bool true  -> let (V_Fun g) = e2 x in g y  (* caml application *)
									 | V_Bool false -> let (V_Fun g) = e3 x in g y (* caml application *)
									 | _ -> raise TypeError                        )))))))
;;

(* ----------------------------------------------------------------- *)

let rec elim = function
  | P_Const c -> P_Const c
  | P_Var x    -> P_Var x


  | P_App (e1, e2) ->  P_App (elim e1, elim e2) 
      
  | P_Abs(s, P_Var s' ) -> 
      if s = s' 
      then  cI 
      else  (P_App (cK, P_Var s'))


  | P_Abs(_, P_Const c) -> P_App (cK, P_Const c )


(* case S_IF *)
  | P_Abs(s,P_App(P_App(P_App(i,e1), P_Abs(_,e2)), P_Abs(_,e3))) when i  == Purify.cIF ->
      P_App (P_App (P_App (cSIF, elim (P_Abs (s,e1))), elim (P_Abs(s,e2))), elim (P_Abs (s, e3)))
	(* see my PhD page 116/117 *)

(* case S_2_IF *)
(* if this case is ommitted, the S_IF case might not be matched when the conditional is under two lambdas *)
  | P_Abs(x,P_Abs(y,P_App(P_App(P_App(i,e1), P_Abs(_,e2)), P_Abs(_,e3)))) when i  == Purify.cIF ->
      let e1' = P_Abs(x,P_Abs (y,e1))
      and e2' = P_Abs(x,P_Abs (y,e2))
      and e3' = P_Abs(x,P_Abs (y,e3)) in
      P_App (P_App (P_App (cS_2_IF, elim e1'), elim e2'), elim e3')

(* case S_2_2 *)
  | P_Abs(x, P_Abs (y, P_App (P_App(e1,e2), e3))) -> 
      let e1' = P_Abs (x, P_Abs(y, e1)) 
      and e2' = P_Abs (x, P_Abs(y, e2)) 
      and e3' = P_Abs (x, P_Abs(y, e3)) 
      in P_App (P_App (P_App (cS_2_2,  elim e1'), elim e2' ), elim e3')
	   
(* case S_2_1 *)
  | P_Abs(x, P_Abs (y, P_App (e1,e2))) ->
      let e1' = P_Abs (x, P_Abs(y, e1))
      and e2' = P_Abs (x, P_Abs(y, e2))
      in P_App( P_App(cS_2_1, elim e1'), elim e2')
	   
(* case S_1_2 *)
  | P_Abs (x,P_App (P_App (e1,e2), e3)) ->
      let e1' = P_Abs (x, e1) 
      and e2' = P_Abs (x, e2) 
      and e3' = P_Abs (x, e3) 
      in P_App (P_App (P_App (cS_1_2,  elim e1'), elim e2' ), elim e3')
    
(* case S_1_1 = case S *)
  | P_Abs(s,P_App(e1,e2)) ->
       P_App(P_App(cS, elim (P_Abs(s,e1))), elim (P_Abs(s,e2)))
      
  | P_Abs(s, (P_Abs(_,_) as t))->
      
       elim (P_Abs(s,elim t))

  ;;

let trans e = check_no_var (elim e) ;;
