(* file B12_B21_SIF2_preauto_trad.ml *)
(* contains SKIBCN algorithm + S_IF combinator + combinators B12... B21... S22 SIF2 (not K2) for multiple abstractions/applications *)
(* auto pre-evaluation of some partial applications when possible *)

open Term ;;
let _ = print_endline "B12_B21_SIF2_preauto";;

(* ----------------------------------------------------------------- *)
(* ---------------------   combinators  ---------------------------- *)
(* ----------------------------------------------------------------- *)

(* combinators S, K and I *)


let cI = P_Const (V_Fun (fun x -> x)) 
and cK = P_Const (V_Fun (fun c -> V_Fun (fun _ -> c)) )
and cS = P_Const (V_Fun ( 
  fun (V_Fun f) -> V_Fun (
    fun (V_Fun g) -> V_Fun ( 
      fun x -> match (f x) with     (* 1 caml capplication *) 
          V_Fun h -> h (g x)     )))) (* 2 caml applications *) 

(* combinator S_IF *)

(* version initiale *)
(*let cSIF =  P_Const (V_Fun (fun (V_Fun e1) -> 
                       V_Fun (fun (V_Fun e2) ->
                                            V_Fun (fun (V_Fun e3) ->
                                                     V_Fun (fun x ->
  match e1 x with (* caml application *)
      V_Bool true  -> e2 x  (* caml application *)
    | V_Bool false -> e3 x  (* caml application *)
    | _ -> raise TypeError                        )))))

*)

(* version compactée + unwrap *)
(*and cSIF =  P_Const (V_Fun (fun (V_Fun e_if) ->
			      V_Fun (fun e_then ->
                                       V_Fun (fun e_else ->
						V_Fun (fun x ->
							 let branch = (
							   match (e_if) x with (* caml application *)
							     | V_Bool true -> e_then
							     | V_Bool false -> e_else
							     | _ -> raise TypeError
							 ) in (unwrap branch) x
						      )
					     )
				    )
			   )
		    )
*)

(* version compactée, sans unwrap *)
(* C'est celle qui donne de meilleurs résultats pour ce combinateur (test fibo) *)

and cSIF =  P_Const (V_Fun (fun (V_Fun e_if) ->
			      V_Fun (fun e_then ->
                                       V_Fun (fun e_else ->
						V_Fun (fun x ->
							 let branch = (
							   match e_if x with (* caml application *)
							     | V_Bool true -> e_then
							     | V_Bool false -> e_else
							     | _ -> raise TypeError
							 ) in   (match  branch with V_Fun f -> f) x 
						      )
					     )
				    )
			   )
		    )



(* combinator for two abstractions on a conditional *)

(* version initiale *)										    
and cS_2_IF = P_Const (V_Fun (fun (V_Fun e1) ->  
                       V_Fun (fun (V_Fun e2) ->
                                            V_Fun (fun (V_Fun e3) ->
                                                     V_Fun (fun x -> 
							      V_Fun (fun y ->
								       let (V_Fun f) = e1 x in match f y with 
									 | V_Bool true  -> let (V_Fun g) = e2 x in g y  (* caml application *)
									 | V_Bool false -> let (V_Fun g) = e3 x in g y (* caml application *)
									 | _ -> raise TypeError                        ))))))


(* version compactée + unwrap *)
(*
and cS_2_IF = P_Const (V_Fun (fun (V_Fun e1) ->                V_Fun (fun e_then ->
                                        V_Fun (fun e_else ->
                                                 V_Fun (fun x ->
                              V_Fun (fun y ->
                                   let (V_Fun f) = e1 x in
                                                                       let branch = match f y with
                                     | V_Bool true  -> e_then
                                     | V_Bool false -> e_else
                                     | _ -> raise TypeError          
                           in  (unwrap ((unwrap branch) x) y)
                                                                    ))))))
 
*)

(* version compactée sans unwrap *)

(*and cS_2_IF = P_Const (V_Fun (fun (V_Fun f_if) -> 
				V_Fun (fun e_then -> 
                                         V_Fun (fun e_else ->
                                                  V_Fun (fun x ->
							   V_Fun (fun y ->
								    let (V_Fun f_if2) = f_if x in
                                                                    let (V_Fun f)  = (match f_if2 y with
											| V_Bool true  -> e_then
											| V_Bool false -> e_else
											| _ -> raise TypeError   )                                 
								    in  
								    let (V_Fun g) = f x     in     g y
												    
                                                                 ))))))*)



(* combinators B C N *)

and cYN = P_Const (V_Fun (fun (V_Fun f) ->
		       V_Fun (fun e ->
				 V_Fun ( fun x ->
					   match  (f x) with V_Fun (h) -> (* caml application *)
					   h e ))))                      (* caml application *)

and cNY = P_Const (V_Fun (fun (V_Fun f) ->
			    V_Fun (fun (V_Fun g) ->
				     V_Fun ( fun x -> f (g x) )))) (* 2 caml applications *)

and cNN =  P_Const (V_Fun (fun (V_Fun f) ->
		       V_Fun (fun e ->
				 V_Fun ( fun _ ->
					   f e))))  (* caml application *)
(*and cYY = cS *)


(* S_2_2 : two abstractions on two applications *)
and cS_2_2 = P_Const (V_Fun (fun (V_Fun f) -> 
		   V_Fun (fun (V_Fun g) ->
			    V_Fun (fun (V_Fun h) ->
				     V_Fun (fun x ->
					      V_Fun (fun y -> 
(*						       let (V_Fun f', V_Fun g', V_Fun h') =  (f x, g x, h x) in
						       let (V_Fun f'', g'', h'') = (f' y, g' y, h' y) in
						       let (V_Fun f''') = f'' g'' in f''' h'')*)
						       
						(*       let (V_Fun f')  = f  x and (V_Fun g') = g  x and V_Fun h'  = h  x in
						       let (V_Fun f'') = f' y and        g'' = g' y and       h'' = h' y in
						       let (V_Fun f''') = f'' g''
						       in f''' h'')  *)
						
						       let V_Fun f'' = (match f x with V_Fun f' -> f') y
						       and       g'' = (match g x with V_Fun g' -> g') y
						       and       h'' = (match h x with V_Fun h' -> h') y
						       in (match f'' g'' with V_Fun f''' -> f''') h''
						    )

					   ))))) 





(* combinators for conditional distribution on 2 applications *)

and cNNN = P_Const (V_Fun (
  fun (V_Fun f) ->
    V_Fun (fun a ->
	     V_Fun (fun b ->
		      V_Fun (fun _ ->
			       (let (V_Fun f') = f a in f') b)))) )
 

(*and cNNY = P_Const (V_Fun (
  fun (V_Fun f) ->
    V_Fun (fun a ->
	     V_Fun (fun (V_Fun h) ->
		      V_Fun (fun x ->
			        (* beware of the order between (f a) and (h x) ?? *)
			       let b = h x
			       in
				 (let (V_Fun f') = f a in f') b ))))) *)

and cNNY = P_Const (V_Fun (
  fun (V_Fun f) ->
    V_Fun (fun a ->
	     V_Fun (fun (V_Fun h) ->
		      V_Fun (fun x ->
			        (* beware of the order between (f a) and (h x) ?? SWITCHED HERE*)
			       (let (V_Fun f') = f a in f') (h x) )))))
and

(* cNYN =  P_Const (V_Fun (
  fun (V_Fun f) ->
    V_Fun (fun (V_Fun g) ->
	     V_Fun (fun b ->
		      V_Fun (fun x ->
			       let a = g x in
			       (let (V_Fun f') = f a in f') b ))))) *)
 cNYN =  P_Const (V_Fun (
  fun (V_Fun f) ->
    V_Fun (fun (V_Fun g) ->
	     V_Fun (fun b ->
		      V_Fun (fun x ->
			       
			       (let (V_Fun f') = f (g x) in f') b )))))


(* and cNYY =  P_Const (V_Fun (
  fun (V_Fun f) ->
    V_Fun (fun (V_Fun g) ->
	     V_Fun (fun (V_Fun h) ->
		      V_Fun (fun x ->
			       let b = h x
			       in let a = g x
			       in (let (V_Fun f') =f a in f') b))))) *)
and cNYY =  P_Const (V_Fun (
  fun (V_Fun f) ->
    V_Fun (fun (V_Fun g) ->
	     V_Fun (fun (V_Fun h) ->
		      V_Fun (fun x ->
			       let b = h x
			       in ( let (V_Fun f') =f (g x) in f') b)))))


and cYNN =  P_Const (V_Fun (
  fun (V_Fun f) ->
    V_Fun (fun a ->
	     V_Fun (fun b ->
		      V_Fun (fun x ->
			       
			       (let (V_Fun f') = f x in  let (V_Fun f'') = f' a in f'') b)))))


and cYNY =  P_Const (V_Fun (
  fun (V_Fun f) ->
    V_Fun (fun a ->
	     V_Fun (fun (V_Fun h) ->
		      V_Fun (fun x ->
			       let b = h x
			       in ( let (V_Fun f') = f x
				    in let (V_Fun f'') = f' a in f'') b )))))


and cYYN =  P_Const (V_Fun (
  fun (V_Fun f) ->
    V_Fun (fun (V_Fun g) ->
	     V_Fun (fun b ->
		      V_Fun (fun x ->
			       let a = g x 
			       in (let (V_Fun f') = f x
				   in  let (V_Fun f'') = f' a in f'') b )))))


and cYYY =  P_Const (V_Fun (
  fun (V_Fun f) ->
    V_Fun (fun (V_Fun g) ->
	     V_Fun (fun (V_Fun h) ->
		      V_Fun (fun x ->
			       let b = h x
			       in 
				 (let a = g x in
				  let (V_Fun f') = f x in let (V_Fun f'') = f' a in f'') b )))))




(* combinators for conditional distribution of two abstractions on one application *)
(*   xx yy                              e1               e2                                                                      *)
and cNN_NN = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun b         -> V_Fun (fun _ -> V_Fun (fun _ -> 
											 (* (e1 e2) *)
											 f b )))))
and cNN_NY = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun (V_Fun g) -> V_Fun (fun _ -> V_Fun (fun y -> 
											 (* e1 (e2 y) *) 
											 f (g y) )))))
and cNN_YN = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun b         -> V_Fun (fun _ -> V_Fun (fun y -> 
											 (* (e1 y) e2 *) 
											 let (V_Fun f') = f y in f' b )))))
and cNN_YY = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun (V_Fun g) -> V_Fun (fun _ -> V_Fun (fun y -> 
											 (* (e1 y) (e2 y) *) 
											 let b= g y in let (V_Fun f') = f y in f' b )))))
and cNY_NN = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun (V_Fun g) -> V_Fun (fun x -> V_Fun (fun _ -> 
											 (*e1 (e2 x) *) 
											 let b= g x in f b )))))
and cNY_NY = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun (V_Fun g) -> V_Fun (fun x -> V_Fun (fun y -> 
											 (* e1 ((e2 x) y) *)
											 let (V_Fun g') = g x in  f (g' y) )))))
and cNY_YN = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun (V_Fun g) -> V_Fun (fun x -> V_Fun (fun y -> 
											 (* ((e1 y) (e2 x)) *) 
											 let b = g x in let (V_Fun f') = f y in f' b)))))
and cNY_YY = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun (V_Fun g) -> V_Fun (fun x -> V_Fun (fun y -> 
											 (* ((e1 y) ((e2 x) y)) *) 
											 let (V_Fun g') = g x in let b = g' y in let (V_Fun f') = f y in f' b)))))
and cYN_NN = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun b         -> V_Fun (fun x -> V_Fun (fun _ -> 
											 (* ((e1 x) e2) *) 
											 let (V_Fun f') = f x in f' b )))))
and cYN_NY = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun (V_Fun g) -> V_Fun (fun x -> V_Fun (fun y ->
											 (* ((e1 x) (e2 y)) *) 
											 let b = g y in let (V_Fun f') = f x in f' b )))))
and cYN_YN = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun b         -> V_Fun (fun x -> V_Fun (fun y ->
											 (* (((e1 x) y) e2) *) 
											 let (V_Fun f') = f x in let (V_Fun f'') = f' y in f'' b )))))
and cYN_YY = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun (V_Fun g) -> V_Fun (fun x -> V_Fun (fun y -> 
											 (* (((e1 x) y) (e2 y)) *)
											 let b = g y in let (V_Fun f') = f x in let (V_Fun f'') = f' y in f'' b)))))
and cYY_NN = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun (V_Fun g) -> V_Fun (fun x -> V_Fun (fun _ ->
											 (* ((e1 x) (e2 x)) *) 
											 let b = g x in let (V_Fun f') = f x in f' b )))))
and cYY_NY = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun (V_Fun g) -> V_Fun (fun x -> V_Fun (fun y -> 
											 (* ((e1 x) ((e2 x) y)) *) 
											 let (V_Fun g') = g x in let b = g' y in let (V_Fun f') = f x in f' b )))))
and cYY_YN = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun (V_Fun g) -> V_Fun (fun x -> V_Fun (fun y -> 
											 (* (((e1 x) y) (e2 x)) *) 
											 let b = g x in let (V_Fun f') = f x in let (V_Fun f'') = f' y in f'' b)))))
and cYY_YY = P_Const (V_Fun (fun (V_Fun f) -> V_Fun (fun (V_Fun g) -> V_Fun (fun x -> V_Fun (fun y -> 
											 (* (((e1 x) y) ((e2 x) y))*) 
											 let (V_Fun g') = g x in let b = g' y in let (V_Fun f') = f x in let (V_Fun f'') = f' y in  f'' b)))))
;;

(* two abstractions on a constant  (K2) *)
(*let cK_2 = P_Const (V_Fun (fun c -> V_Fun (fun _ -> V_Fun (fun _ -> c)) )) ;;*)

(*let cYY = cS ;;*)


(* ----------------------------------------------------------------- *)


let rec skip x = function
  | P_Const _ -> true
  | P_Var y -> x <> y  (*false*) (* vérifier *) (* la version x<>y donne de meilleurs résultats mais est-elle correcte? *)
  | P_App (e1, e2) -> false
  | P_Abs (y, e) -> x=y || not (free x e) ;; 

  

(* condition d'utilisation de pre_eval_app : la branche gauche de l'application est soit une cosntante fonctionnelle, soit une application . On n'appelle pre_eval_app que lorsqu'on a une application d'un combinateur à des arguments. Et les arguments ont déjà été pre-evalués si possible (on espère). *)

 let rec pre_eval_app e = match e with
  | P_App (e1,e2) ->
      ( match (pre_eval_app e1, pre_eval_app e2) 
	with 
	  | (P_Const (V_Fun f),P_Const v) -> P_Const (f v)
	  | (e'1,e'2) -> P_App (e'1,e'2)
      )
  | _ -> e
      
(* A SUPPRIMER *)
(* two abstractions on a constant  (K2) *)
(*let cK_2 = P_Const (V_Fun (fun c -> V_Fun (fun _ -> V_Fun (fun _ -> c)) ))*)


let rec elim = function

  | P_Const c -> P_Const c
  | P_Var x    -> P_Var x

  | P_App (e1, e2) ->  P_App (elim e1, elim e2) 
      

(* -- Case -- K *)
  | P_Abs(s, P_Var s' ) -> 
      if s = s' 
      then  cI 
      else  (P_App (cK, P_Var s'))

  | P_Abs(_,P_Const c) -> P_Const (V_Fun (fun _ -> c)) (* hand-made pre-eval *)

(* A SUPPRIMER *)
(* K2 automatic pre_eval *)
(*  | P_Abs(_, P_Abs(_, P_Const c))    ->  pre_eval_app (P_App(cK_2, P_Const c)) 

  | P_Abs(x, P_Abs(y, P_Var z)) when (x<>z && y<>z)->  P_App(cK_2, P_Var z) *)

(* case S_IF *)
  | P_Abs(s,P_App(P_App(P_App(i,e1), P_Abs(_,e2)), P_Abs(_,e3))) when i  == Purify.cIF -> 
      pre_eval_app (P_App (P_App (P_App (cSIF, elim (P_Abs (s,e1))), elim (P_Abs(s,e2))), elim (P_Abs (s, e3)))) 
      

(* case S_2_IF *)
  (* if this case is ommitted, the S_IF case might not be matched when the conditional is under two lambdas *)
  | P_Abs(x,P_Abs(y,P_App(P_App(P_App(i,e1), P_Abs(_,e2)), P_Abs(_,e3)))) when i  == Purify.cIF ->
      
      let e1' = P_Abs(x,P_Abs (y,e1))
      and e2' = P_Abs(x,P_Abs (y,e2)) 
      and e3' = P_Abs(x,P_Abs (y,e3)) in
	pre_eval_app (P_App (P_App (P_App (cS_2_IF, elim e1'), elim e2'), elim e3')) 

(* FIXME : B_IF and C_IF and N_IF would be nice too, but is it possible? *)
      

(* case S_2_2 *) (* nb : j'ai essaye de retirer ce cas pour privilegier les variantes de S21 mais c'est plus mauvais *)
  | P_Abs(x, P_Abs (y, P_App (P_App(e1,e2), e3))) 
      (* if possible, we prefer to use cases with skipping *)
      when not (skip x e3 || skip y e3 ) 
      
      -> 
      let e1' = P_Abs (x, P_Abs(y, e1)) 
      and e2' = P_Abs (x, P_Abs(y, e2))  
      and e3' = P_Abs (x, P_Abs(y, e3))  
      in pre_eval_app (P_App (P_App (P_App (cS_2_2,  elim e1'), elim e2' ), elim e3'))


  

(* case S_2_1 *)
  | P_Abs(x, P_Abs (y, P_App (e1,e2))) -> 
	let (c,a1,a2) =
	  (match (skip x e1, skip x e2, skip y e1, skip y e2) with
	     | (true,  true,  true,  true ) -> (cNN_NN, e1         ,           e2                    )
	     | (true,  true,  true,  false) -> (cNN_NY, e1         ,           P_Abs(y,e2)           )
	     | (true,  true,  false, true ) -> (cNN_YN, P_Abs(y,e1),           e2                    )
	     | (true,  true,  false, false) -> (cNN_YY, P_Abs(y,e1),           P_Abs(y,e2)           )
	     | (true,  false, true,  true ) -> (cNY_NN, e1         ,           P_Abs(x,e2)           )
	     | (true,  false, true,  false) -> (cNY_NY, e1         ,           P_Abs(x, P_Abs(y, e2)))
	     | (true,  false, false, true ) -> (cNY_YN, P_Abs(y,e1),           P_Abs(x,e2)           )
	     | (true,  false, false, false) -> (cNY_YY, P_Abs(y,e1),           P_Abs(x, P_Abs(y, e2)))
	     | (false, true,  true,  true ) -> (cYN_NN, P_Abs(x,e1),           e2                    )
	     | (false, true,  true,  false) -> (cYN_NY, P_Abs(x,e1),           P_Abs(y,e2)           )
	     | (false, true,  false, true ) -> (cYN_YN, P_Abs(x, P_Abs(y,e1)), e2                    )
	     | (false, true,  false, false) -> (cYN_YY, P_Abs(x, P_Abs(y,e1)), P_Abs(y,e2)           )
	     | (false, false, true,  true ) -> (cYY_NN, P_Abs(x,e1)          , P_Abs(x,e2)           )           
	     | (false, false, true,  false) -> (cYY_NY, P_Abs(x,e1)          , P_Abs(x, P_Abs(y, e2)))
	     | (false, false, false, true ) -> (cYY_YN, P_Abs(x, P_Abs(y,e1)), P_Abs(x,e2)           )
	     | (false, false, false, false) -> (cYY_YY, P_Abs(x, P_Abs(y,e1)), P_Abs(x, P_Abs(y, e2)))
	)
      in pre_eval_app(P_App(P_App(c, elim a1), elim a2))

	   
(* case S_1_2 *)
  | P_Abs (x,P_App (P_App (e1,e2), e3)) ->  
      
      let e1' = P_Abs (x, e1) 
      and e2' = P_Abs (x, e2) 
      and e3' = P_Abs (x, e3) 
      in 
      let (c, a1, a2, a3) = 
	(match (skip x e1, skip x e2, skip x e3) with
	   | (true , true , true ) -> (cNNN, e1  , e2  , e3  )
	   | (true , true , false) -> (cNNY, e1  , e2  , e3' )
	   | (true , false, true ) -> (cNYN, e1  , e2' , e3  )
	   | (true , false, false) -> (cNYY, e1  , e2' , e3' )
	   | (false, true , true ) -> (cYNN, e1' , e2  , e3  )
	   | (false, true , false) -> (cYNY, e1' , e2  , e3' )
 	   | (false, false, true ) -> (cYYN, e1' , e2' , e3  )
	   | (false, false, false) -> (cYYY, e1' , e2' , e3' )
	)
      in pre_eval_app(P_App(P_App(P_App(c, elim a1), elim a2), elim a3))
	   
	   
   
(* case B, C, N, S *) 
  | P_Abs(s,P_App(e1,e2)) ->
      (match (skip s e1,skip s e2) with
	 | true,  true  -> pre_eval_app( P_App(P_App(cNN, elim e1)           , elim e2)            )
	 | true,  false -> pre_eval_app( P_App(P_App(cNY, elim e1)           , elim (P_Abs(s,e2))) )
	 | false, true  -> pre_eval_app( P_App(P_App(cYN, elim (P_Abs(s,e1))), elim e2)            )
	 | false, false -> pre_eval_app( P_App(P_App(cS , elim (P_Abs(s,e1))), elim (P_Abs(s,e2))) )
      )

  | P_Abs(s, (P_Abs(_,_) as t))-> 
      
       elim (P_Abs(s,elim t)) 

  ;; 
 
let trans e = check_no_var (elim e) ;;
