 {
        open Parser_ho        (* The type token is defined in parser.mli *)
	open Term
        exception Eof
  let plus   = E_Const (V_Fun Externals.plus)
  let minus  = E_Const (V_Fun Externals.minus)
  let inf    = E_Const (V_Fun Externals.inf)
  let equals = E_Const (V_Fun Externals.equals)
  let diff   = E_Const (V_Fun Externals.diff)

  let empty = E_Const (V_Fun Externals.empty)
  let tl    = E_Const (V_Fun Externals.tl)
  let hd    = E_Const (V_Fun Externals.hd)
  let cons  = E_Const (V_Fun Externals.ccons)
  let log_and  = E_Const (V_Fun Externals.log_and)

  let rev =  E_Const (V_Fun Externals.rev)
  let flatten =  E_Const (V_Fun Externals.flatten)
  let length =  E_Const (V_Fun Externals.length)
  let iter =  E_Const (V_Fun Externals.iter)
  let filter =  E_Const (V_Fun Externals.filter)

  let print  = E_Const (V_Fun Externals.print) 

let incr_linenum lexbuf =
    let pos = lexbuf.Lexing.lex_curr_p in
        lexbuf.Lexing.lex_curr_p <- { pos with
	      Lexing.pos_lnum = pos.Lexing.pos_lnum + 1;
              Lexing.pos_bol = pos.Lexing.pos_cnum;
}
	;; (* from a tutorial *)

        }
        rule token = parse
            [' ' '\t']     { token lexbuf }     (* skip blanks *)
          | ['\n' ]        { incr_linenum lexbuf ; token lexbuf }
          | ['0'-'9']+ as lxm { INT(int_of_string lxm) }
	  | ['a'-'z']+ as lxm { match lxm with 
				  | "if"   -> IF 
				  | "then" -> THEN
				  | "else" -> ELSE
				  | "fi"   -> FI
				  | "let"  -> LET
				  | "in"   -> IN
				  | "fun" -> FUN
				  | "begin" -> BEGIN
				  | "end" -> END
				  | "empty" -> CONST(empty)
				  | "tl" -> CONST(tl)
				  | "hd" -> CONST(hd)
				  | "rev" -> CONST(rev)
				  | "print" -> CONST(print)
				  | "flatten" -> CONST(flatten)
				  | "length" -> CONST(length)
				  | "iter"   -> CONST(iter)
				  | "filter"   -> CONST(filter)

				  | _ -> IDENT(lxm) }
          | '+'            { BINOP(plus) } 
	  | "->"           { ARROW }
          | '-'            { BINOP(minus) }
(*          | '*'            { TIMES } *)
(*          | '/'            { DIV } *)
          | '('            { LPAREN }
          | ')'            { RPAREN }
	  | '{'            { LACC }
	  | '}'            { RACC }
	  | '\\'           { LAMBDA  } (* backslash *)
	  | '.'            { DOT    }
	  | ":="           { SET    }
	  | '!'            { GET    }
	  | "<="           { BINOP(inf)}
	  | "=="           { BINOP(equals)}
	  | "<>"           { BINOP(diff)}
	  | '='            { BE }
          | ','            { BINOP(cons) }
	  | "::"           { BINOP(cons) }
	  | "[]"           { EMPTYLIST }
	  | ';'            { SEQUENCE }
	  | "&&"           { BINOP(log_and) }
	  | '&'            { BINOP(log_and) } 
	  | eof            { EOF }
	  
