%{
  open Term ;;

  let plus_h  = E_ExtFun Externals.plus
  let minus_h = E_ExtFun Externals.minus
  let inf_h   = E_ExtFun Externals.inf

  let empty_h = E_ExtFun Externals.empty
  let tl_h    = E_ExtFun Externals.tl
  let hd_h    = E_ExtFun Externals.hd



  let plus_c  = Classical.E_ExtFun Classical.plus
  let minus_c = Classical.E_ExtFun Classical.minus
  let inf_c   = Classical.E_ExtFun Classical.inf

  let empty_c = Classical.E_ExtFun Classical.empty
  let tl_c    = Classical.E_ExtFun Classical.tl
  let hd_c    = Classical.E_ExtFun Classical.hd

  let dummy = ""

%}
        %token <int> INT
        %token <string> IDENT
/*        %token <int> MIDENT*/

        %token PLUS MINUS TIMES DIV INF
        %token IF THEN ELSE FI
        %token LET BE IN 
        %token LPAREN RPAREN LACC RACC
	%token LAMBDA DOT
        %token EOF
        %token CONS
        %token SET GET
        %token SEQUENCE

        %left PLUS MINUS        /* lowest precedence  */
        %left TIMES DIV         /* medium precedence  */
        %nonassoc UMINUS        /* highest precedence */


        %start main               /*  entry point    */
        %start main_c             /*  entry point    */
        %type <Term.expr> main
        %type <Classical.expr> main_c
        %%
        main:
            expr EOF                { $1 }
        ;
        expr:
            INT                                { E_Int $1 }
          | IDENT                              { match $1 with 
						   | "empty" -> empty_h
						   | "tl"    -> tl_h
						   | "hd"    -> hd_h
						       (* this is not very good since then build-in function cannot be masked by local variables *)
						   | _ -> E_Var $1 }
	  | LACC expr RACC                     { $2 }
          | LPAREN expr expr        RPAREN     { E_App ($2, $3) }
          | LPAREN expr PLUS  expr  RPAREN     { E_App (E_App (plus_h , $2), $4) }
          | LPAREN expr MINUS expr  RPAREN     { E_App (E_App (minus_h, $2), $4) }
/*          | LPAREN expr TIMES expr  RPAREN     { E_App (E_App (times_h, $2), $4) }          */
	  | LPAREN expr INF   expr  RPAREN     { E_App (E_App (inf_h, $2), $4) }
	  | LAMBDA IDENT DOT expr              { E_Abs ($2,$4) }
	  | IF expr THEN expr ELSE expr FI     { E_If ($2, $4, $6) }
	      
	  | LPAREN expr CONS expr RPAREN       { E_Cons ($2, $4) }
	 
	  | LET IDENT BE expr IN expr    { E_App (E_Abs($2,$6), $4)}
	  | GET IDENT                    { E_Get ($2)              }
	  | LPAREN IDENT SET expr RPAREN { E_Set ($2, $4)          }
	  | LPAREN expr SEQUENCE expr RPAREN  { E_App (E_Abs (dummy,$4), $2) }
        ;

       
        main_c:
            expr_c EOF                { $1 }
        ;
        expr_c:
            INT                                { Classical.E_Int $1 }
          | IDENT                              { match $1 with 
						   | "empty" -> empty_c 
						   | "tl"    -> tl_c
						   | "hd"    -> hd_c
						   | _ -> Classical.E_Var $1 }
	  | LACC expr_c RACC                     { $2 }
          | LPAREN expr_c expr_c        RPAREN     { Classical.E_App ($2, $3) }
          | LPAREN expr_c PLUS  expr_c  RPAREN     { Classical.E_App (Classical.E_App (plus_c , $2), $4) }
          | LPAREN expr_c MINUS expr_c  RPAREN     { Classical.E_App (Classical.E_App (minus_c, $2), $4) }
/*        | LPAREN expr_c TIMES expr_c  RPAREN     { Classical.E_App (Classical.E_App (times_c, $2), $4) }          */
	  | LPAREN expr_c INF   expr_c  RPAREN     { Classical.E_App (Classical.E_App (inf_c,   $2), $4) }
	  | LAMBDA IDENT DOT expr_c              { Classical.E_Abs ($2,$4) }
	  | IF expr_c THEN expr_c ELSE expr_c FI     { Classical.E_If ($2, $4, $6) }

	  | LPAREN expr_c CONS expr_c RPAREN       { Classical.E_Cons ($2, $4) }

	  | LET IDENT BE expr_c IN expr_c {Classical.E_App (Classical.E_Abs($2,$6), $4)}
	  | GET IDENT                    { Classical.E_Get ($2)              }
	  | LPAREN IDENT SET expr_c RPAREN { Classical.E_Set ($2, $4)          }
 	  | LPAREN expr_c SEQUENCE expr_c RPAREN  { Classical.E_App (Classical.E_Abs (dummy,$4), $2) }

        ;

