%{
  exception SyntaxError of string * string 

  open Classical ;;

  let dummy = ""

let string_of_position p =
     "line " ^ (string_of_int p.Lexing.pos_lnum) ^
     ", character " ^ (string_of_int p.Lexing.pos_cnum)

%}


        %token MIN
        %token <int> INT
        %token <string> IDENT
        %token <Classical.expr > BINOP
        %token <Classical.expr> CONST
/*        %token <int> MIDENT*/

        %token IF THEN ELSE FI
        %token LET BE IN 
        %token LPAREN RPAREN LACC RACC
	%token LAMBDA DOT FUN ARROW
        %token EOF
        %token CONS EMPTYLIST
        %token SET GET
        %token SEQUENCE
        %token BEGIN END
%token TOP 


        %left PLUS MINUS        /* lowest precedence  */
        %left TIMES DIV         /* medium precedence  */
        %nonassoc UMINUS        /* highest precedence */


        %start main               /*  entry point    */
        %type <Classical.expr> main
        %%
        main:
            expr EOF                { $1 }
        ;

  expr: 
  | simpleexpr              {$1}
  | expr simpleexpr         {E_App ($1,$2)}

  | expr BINOP    simpleexpr     { E_App (E_App($2,$1), $3) }
  | expr SEQUENCE simpleexpr   { E_App (E_Abs (dummy,$3), $1) } 
    /*          | expr TIMES simpleexpr      { E_App (E_App (times, $1), $3) }          */

  | error %prec MIN{ let p1 = (Parsing.rhs_start_pos 1) 
            and p2 = (Parsing.symbol_end_pos()) in
            raise (SyntaxError(string_of_position p1, string_of_position p2)) }



;

  simpleexpr:
  | LPAREN expr RPAREN                     { $2 }
  | BEGIN expr END { $2 }
  | INT                                { E_Const (V_Int $1) }
  | CONST                              { $1 }
  | EMPTYLIST                          { E_Const (V_List [])}
  | IDENT                              { E_Var $1 }
      
  | LET IDENT BE expr IN expr    { E_App (E_Abs($2,$6), $4)} 
  | GET IDENT                    { E_Get ($2)              }
  | IDENT SET expr  { E_Set ($1, $3)          } 
  | IF expr THEN expr ELSE expr FI     { E_If ($2, $4, $6) }      
  | FUN    IDENT ARROW expr              { E_Abs ($2,$4) } 
  | LAMBDA IDENT DOT   expr              { E_Abs ($2,$4) } 

  ;


/* ancienne version */
/*
        %token <int> INT
        %token <string> IDENT
        %token <int> MIDENT

        %token PLUS MINUS TIMES DIV INF
        %token IF THEN ELSE FI
        %token LET BE IN 
        %token LPAREN RPAREN LACC RACC
	%token LAMBDA DOT
        %token EOF
        %token CONS EMPTYLIST
        %token SET GET
        %token SEQUENCE

        %left PLUS MINUS   
        %left TIMES DIV    
        %nonassoc UMINUS   


        %start main_c      
        %type <Classical.expr> main_c
        %%
 
       
        main_c:
            expr_c EOF                { $1 }
        ;
        expr_c:
            INT                                { E_Const(V_Int $1) }
          | EMPTYLIST                          { E_Const (V_List [])}
          | IDENT                              { match $1 with 
						   | "empty" -> empty 
						   | "tl"    -> tl
						   | "hd"    -> hd
						   | _ -> E_Var $1 }
	  | LACC expr_c RACC                     { $2 }
          | LPAREN expr_c expr_c        RPAREN     { E_App ($2, $3) }
          | LPAREN expr_c PLUS  expr_c  RPAREN     { E_App (E_App (plus , $2), $4) }
          | LPAREN expr_c MINUS expr_c  RPAREN     { E_App (E_App (minus, $2), $4) }

	  | LPAREN expr_c INF   expr_c  RPAREN     { E_App (E_App (inf,   $2), $4) }
	  | LAMBDA IDENT DOT expr_c              { E_Abs ($2,$4) }
	  | IF expr_c THEN expr_c ELSE expr_c FI     { E_If ($2, $4, $6) }

	  | LPAREN expr_c CONS expr_c RPAREN       { E_App(E_App(cons ,$2), $4) }

	  | LET IDENT BE expr_c IN expr_c { E_App (E_Abs($2,$6), $4)}
	  | GET IDENT                    { E_Get ($2)              }
	  | LPAREN IDENT SET expr_c RPAREN { E_Set ($2, $4)          }
 	  | LPAREN expr_c SEQUENCE expr_c RPAREN  { E_App (E_Abs (dummy,$4), $2) }

        ;
*/
