%{
(*  exception SyntaxError of Lexing.position * Lexing.position * string*)
  exception SyntaxError of string * string 
  open Term ;;

  let dummy = ""

let string_of_position p =
     "line " ^ (string_of_int p.Lexing.pos_lnum) ^
     ", character " ^ (string_of_int p.Lexing.pos_cnum)

%}
        %token MIN
        %token <int> INT
        %token <string> IDENT
        %token < Term.expr > BINOP
        %token <Term.expr> CONST
/*        %token <int> MIDENT*/


        %token IF THEN ELSE FI
        %token LET BE IN 
        %token LPAREN RPAREN LACC RACC
	%token LAMBDA DOT FUN ARROW
        %token EOF
        %token CONS EMPTYLIST
        %token SET GET
        %token SEQUENCE
        %token BEGIN END
%token TOP 


        %left PLUS MINUS        /* lowest precedence  */
        %left TIMES DIV         /* medium precedence  */
        %nonassoc UMINUS        /* highest precedence */


        %start main               /*  entry point    */
        %type <Term.expr> main
        %%
        main:
            expr EOF                { $1 }
        ;

  expr: 
  | simpleexpr              {$1}
  | expr simpleexpr         {E_App ($1,$2)}

  | expr BINOP    simpleexpr     { E_App (E_App($2,$1), $3) }
  | expr SEQUENCE simpleexpr   { E_App (E_Abs (dummy,$3), $1) } 
    /*          | expr TIMES simpleexpr      { E_App (E_App (times, $1), $3) }          */

  | error %prec MIN{ let p1 = (Parsing.rhs_start_pos 1) 
            and p2 = (Parsing.symbol_end_pos()) in
            raise (SyntaxError(string_of_position p1, string_of_position p2)) }



;

  simpleexpr:
  | LPAREN expr RPAREN                     { $2 }
  | BEGIN expr END { $2 }
  | INT                                { E_Const (V_Int $1) }
  | CONST                              { $1 }
  | EMPTYLIST                          { E_Const (V_List [])}
  | IDENT                              { E_Var $1 }
      
  | LET IDENT BE expr IN expr    { E_App (E_Abs($2,$6), $4)} 
  | GET IDENT                    { E_Get ($2)              }
  | IDENT SET expr  { E_Set ($1, $3)          } 
  | IF expr THEN expr ELSE expr FI     { E_If ($2, $4, $6) }      
  | FUN    IDENT ARROW expr              { E_Abs ($2,$4) } 
  | LAMBDA IDENT DOT   expr              { E_Abs ($2,$4) } 

  ;
