type ident = string ;;



(* ------- *)


(* ------------------------------- *)
(* Values : result of interpreting *)
(* ------------------------------- *)

type value = 
  | V_Int of int
  | V_Bool of bool
  | V_Fun of (value -> value)
(* no variables, no closures, no applications *)

  | V_List of value list
(*  | String of string *)

;;


exception TypeError ;;

(*let unwrap = function (V_Fun f) -> f ;;*)


(* ----------------------------------- *)
(* the type of expressions to evaluate *)
(* ----------------------------------- *)

type expr = 
  | E_Const of value
(*  | E_Int of int
  | E_Bool of bool *)
  | E_Var of ident
  | E_Abs of ident * expr
  | E_App of expr * expr
(*  | E_ExtFun of (value->value)*)
  | E_If of expr * expr * expr

(*  | E_Cons of expr * expr *)

  | E_Get of Store.tag
  | E_Set of Store.tag * expr

;;




(* ----------------------------------------- *)
(* the type of purely functional expressions *)
(* ----------------------------------------- *)
type p_expr = 
  | P_Const of value
  | P_Var of ident
  | P_Abs of ident * p_expr
  | P_App of p_expr * p_expr
  (* no If *)
  (* no closure *) 


;;


(* --------------------------------------- *)
(* expressions after variable elimination. *)
(* all functions are caml functions        *)
(* --------------------------------------- *)
type  m_expr = 
  | M_Const of value
  | M_App of  m_expr *  m_expr
;;



(* -------------------- *)
(* Printers             *)
(* -------------------- *)


let rec string_of_value = function
  | V_Int i -> string_of_int i
  | V_Bool true -> "true"
  | V_Bool false -> "false"
  | V_Fun f -> "fun"

  | V_List l -> "[ " ^ List.fold_right (fun x-> fun r -> string_of_value x ^ " , " ^ r) l " ]"

(*  | String s  -> s *)

;; 
(*
let rec string_of_m = function
  | M_Const v -> string_of_value v
  | M_App (e1,e2) -> "( " ^ string_of_m e1 ^ " " ^ string_of_m e2 ^ " )" ;;
*)

(* ------------------------- *)
(* Useful functions on terms *)
(* ------------------------- *)

exception UnexpectedError

(*exception FreeVar of string
exception LambdaAbstraction*)

let rec free x = function
  | P_Const _ -> false
  | P_Var y  -> x = y 
  | P_Abs (y, e)  -> x<>y && free x e
  | P_App (e1,e2) -> free x e1 || free x e2

let rec check_no_var = function
  | P_Const c -> M_Const c
  | P_App (e1, e2) -> M_App (check_no_var e1, check_no_var e2)
  | P_Var x -> raise UnexpectedError (*raise (FreeVar x)*)
  | P_Abs _ -> raise UnexpectedError (* LambdaAbstraction *)

;;

(*let rec freevars l = function
  | E_Const _ -> []
  | E_App (e1,e2) -> freevars l e1 @ freevars l e2
  | E_Var x -> if List.mem x l then [] else [x]
  | E_Set (_,e) -> freevars l e
  | E_Get _ -> []
  | E_Abs (x,e) -> freevars (x::l) e
  | E_If (a,b,c) -> freevars l a @ freevars l b @ freevars l c*)

let rec freevars l = function
  | E_Const _ -> false
  | E_App (e1,e2) -> freevars l e1 || freevars l e2
  | E_Var x -> not (List.mem x l )
  | E_Set (_,e) -> freevars l e
  | E_Get _ -> false
  | E_Abs (x,e) -> freevars (x::l) e
  | E_If (a,b,c) -> freevars l a || freevars l b || freevars l c

